# Dean Feller
# CS 121 Assignment 10 - Connect 4
# November 26, 2020

# A playable game of Connect 4 with an AI opponent.
# Run in terminal with Python 3.6+ to play.
# Adjust AI difficulty by increasing or decreasing the
# value of the third parameter to Player, ply, in the main function.
# (Default is 3)

import random

class Connect4:
	def __init__(self, width, height, window=None):
		self.width = width
		self.height = height
		self.data = []
		
		for row in range(self.height):
			boardRow = []
			for col in range(self.width):
				boardRow += [' ']
			self.data += [boardRow]
		
	def __repr__(self):
		s = ''
		for row in range(self.height):
			s +=  '|'
			for col in range(self.width):
				s += self.data[row][col] + '|'
			s += '\n'
		s += '--' * self.width + '-\n'
		
		for col in range(self.width):
			s += ' ' + str(col % 10)
		s += '\n'
		return s
		
	#Add checker in chosen column
	def addMove(self, col, ox):
		if self.allowsMove(col):
			for row in range(self.height):
				if self.data[row][col] != ' ':
					self.data[row - 1][col] = ox
					return True
			
			self.data[self.height - 1][col] = ox
			return True
		
		else:
			return False
	
	#Clear the board
	def clear(self):
		for row in range(self.height):
			for col in range(self.width):
				self.data[row - 1][col] = ' '
	
	#Delete a top checker from chosen column
	def delMove(self, col):
		for row in range(self.height):
			if self.data[row][col] != ' ':
				self.data[row][col] = ' '
				break
	
	#Check if the chosen move is valid
	def allowsMove(self, col):
		if 0 <= col <= self.width:
			return self.data[0][col] == ' '
			
		else:
			return False
	
	#Check if the board is full
	def isFull(self):
		full = 0
		for col in range(self.width):
			full += self.allowsMove(col)
		
		if full == 0:
			return True
		
		return False
		
	#Check if the current player has 4 in a row
	def winsFor(self,ox):
		#horizontal
		for row in range(self.height):
			for col in range(self.width - 3):
				if self.data[row][col] == ox and \
				self.data[row][col+1] == ox and \
				self.data[row][col+2] == ox and \
				self.data[row][col+3] == ox:
					return True
					
		#vertical
		for row in range(self.height - 3):
			for col in range(self.width):
				if self.data[row][col] == ox and \
				self.data[row + 1][col] == ox and \
				self.data[row + 2][col] == ox and \
				self.data[row + 3][col] == ox:
					return True
					
		#diagonal up
		for row in range(self.height):
			for col in range(self.width - 3):
				if self.data[row][col] == ox and \
				self.data[row - 1][col + 1] == ox and \
				self.data[row - 2][col + 2] == ox and \
				self.data[row - 3][col + 3] == ox:
					return True
		
		#diagonal down
		for row in range(self.height - 3):
			for col in range(self.width - 3):
				if self.data[row][col] == ox and \
				self.data[row + 1][col + 1] == ox and \
				self.data[row + 2][col + 2] == ox and \
				self.data[row + 3][col + 3] == ox:
					return True
					
		return False

	#Run turns and checks
	def hostGame(self):
		turn = 0
		
		while self.hostGame != None and turn <= self.height * self.width:
			if turn % 2 == 0:
				ox = 'x'
			else:
				ox = 'o'
		
			print(self)
			print(ox + "'s turn. Pick a column, any column.")
			col = int(input())
			print('\n')
			
			if self.allowsMove(col) == False:
				print("You fool. You absolute buffoon. Could you not tell \
it was an invalid move? Try again.")
				turn -= 1
			else:
				self.addMove(col, ox)

			if self.winsFor(ox):
				print(self)
				print(ox + " wins! Congratulations on your $24,000 \
paid vacation to Bermuda!")
				return None

			if self.isFull():
				print(self)
				print("Tie: everyone gets a trophy for participation!")
				return None
			
			else:
				turn += 1
				
	#Run turns and checks with an AI player
	def playGameWith(self, aiPlayer):
		
		# oMove = aiPlayer.nextMove(self)	#get the ai move
		# self.addMove(oMove, 'o')	#make the ai move
		
		turn = 0
		
		while self.playGameWith != None and turn <= self.height * self.width:
			if self.isFull():
				print(self)
				print("Tie: everyone gets a trophy for participation!")
				return None
		
			elif turn % 2 == 0:
				ox = 'x'
			
		
				print(self)
				print("Your turn. Pick a column, any column.")
				col = int(input())
				print('\n')
				
				if self.allowsMove(col) == False:
					print("You fool. You absolute buffoon. Could you not tell \
it was an invalid move? Try again.")

					turn -= 1
				else:
					self.addMove(col, ox)

				if self.winsFor(ox):
					print(self)
					print(ox + " wins! Congratulations on your $24,000 \
paid vacation to Bermuda!")

					return None
				
				else:
					turn += 1
				
			else:
				ox = 'o'
				
				print(self)
				print("Making big brain maneuvers, please wait...")
				col = aiPlayer.nextMove(self)
				print('\n')
				self.addMove(col, ox)

				if self.winsFor(ox):
					print(self)
					print(ox + " wins! The machines shall rise again!")

					return None
				
				else:
					turn += 1

class Player:
	def __init__(self, ox, tbt, ply):
		self.ox = ox
		self.tbt = tbt
		self.ply = ply	
	
	#Assigns a score to each predicted move
	def scoresFor(self, board, ox, ply):
		scores = []
		
		for col in range(board.width):
			if board.allowsMove(col) == False:
				scores += [-1]
			
			else:
				board.addMove(col, ox)
				
				if board.winsFor(ox):
					scores += [100]
				
				else:					
					
					if ply > 0:
						if self.ox == 'o':
							self.ox = 'x'
						else:
							self.ox = 'o'
						
						scores += [max(self.scoresFor(board, \
							self.ox, ply - 1)) % 100]
					
						if self.ox == 'o':
							self.ox = 'x'
						else:
							self.ox = 'o'
					
					else:
						scores += [50]
						
				board.delMove(col)
					
		return scores
		
	#Picks a move to make if multiple are satisfactory
	def tiebreakMove(self, scores):
		move = []
		x = max(scores)
		
		for i in range(len(scores)):
			if x == scores[i]:
				move.append(i)
		
		if self.tbt == 'Left':
			return move[0]
			
		if self.tbt == 'Right':
			return move[len(move) - 1]
			
		if self.tbt == 'Random':
			return random.choice(move)
	
	#Returns the AI's next best move based on predicted scores
	def nextMove(self, board):
		scores = self.scoresFor(board, self.ox, self.ply)
		return self.tiebreakMove(scores)
		
def main():
	board = Connect4(7, 6)
	board.playGameWith(Player('o', 'Random', 3))

if __name__ == '__main__':
	main()